# -*- coding: utf-8 -*-
# filename: spd.py
# Copyright 2017 Mario Gutiérrez-Roig <mariogutierrezroig@gmail.com>
# Copyright 2020 Stefano Costa <steko@iosa.it>
#
# This file is part of IOSACal, the IOSA Radiocarbon Calibration Library.

# IOSACal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# IOSACal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with IOSACal.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from iosacal.core import CalibrationCurve, RadiocarbonDetermination


def spdsum(spdlist, norm=True):
    """Sums several SPDs stored in a list.

    Arguments:
        norm -- If 'True' final SPD will be normalized.

    """

    spd = spdlist[0].copy()

    maxi = np.max([np.max(np.array(x).T[0]) for x in spdlist])
    mini = np.min([np.min(np.array(x).T[0]) for x in spdlist])

    # Creates a void vector where perform the sum
    xrange = np.arange(mini, maxi + 1, 1)
    yrange = np.zeros((int)(maxi) + 1 - (int)(mini))

    for d in spdlist:
        # Reshapes every array by adding zeros at the head and tail
        y = np.lib.pad(
            d.T[1],
            (np.int(d.T[0][0] - mini), np.int(maxi - d.T[0][-1])),
            "constant",
            constant_values=0,
        )

        # Summing over all histogram
        yrange += y

    # Normalizating the SPD in calendar scale
    if norm == True:
        yrange = yrange / np.sum(yrange)

    spd.resize(len(xrange), 2, refcheck=False)

    spd.T[0] = xrange
    spd.T[1] = yrange

    spd.ndates = np.sum([x.ndates for x in spdlist])

    return spd


class SPD(np.ndarray):
    """A Sum of Probability Distributions of calibrated dates"""

    def __new__(cls, caldates, norm=True):
        """Initializes the SPD class from calibrated dates

        Arguments:
            caldates -- List with all calibrated dates (after R.calibrate()) in a list
            norm -- If 'True' SPD will be normalized

        """

        # Finds maximum and minimum value of calibrated dates
        maxi = np.max([np.max(np.array(x).T[0]) for x in caldates])
        mini = np.min([np.min(np.array(x).T[0]) for x in caldates])

        # Creates a void vector where perform the sum
        xrange = np.arange(mini, maxi + 1, 1)
        yrange = np.zeros((int)(maxi) + 1 - (int)(mini))

        for d in caldates:
            # Disentangles and reverses the arrays
            x = np.array(d).T[0][::-1]
            y = np.array(d).T[1][::-1]

            # Reshapes every array by adding zeros at the head and tail
            yy = np.lib.pad(
                y,
                (np.int(x[0] - mini), np.int(maxi - x[-1])),
                "constant",
                constant_values=0,
            )

            # Summing over all histogram
            yrange += yy

        # Normalizating the SPD in calendar scale
        if norm == True:
            yrange = yrange / np.sum(yrange)

        obj = np.asarray(np.column_stack((xrange, yrange))).view(cls)
        obj.ndates = len(caldates)

        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
        self.ndates = getattr(obj, "ndates", None)

    def normalize(self):
        """Normalizes the SPD"""

        y = np.array(self.T[1])

        self.T[1] /= np.sum(y)

    def taphcorr(self, function="Surovell", factor=1):
        """Performs a taphonomic correction over the SPD

        Taphonomic Correction can be made according two different curves
        indicated by the variable 'function': Surovell's or Williams' correction:

        Surovell's correction:

            n(t) = 5.726442 * 10^6 * (t + 2176.4)^-1.3925309

        Surovell, T. A., Finley, J. B., Smith, G. M., Brantingham, P. J.,
        & Kelly, R. (2009). Correcting temporal frequency distributions for
        taphonomic bias. Journal of Archaeological Science, 36(8), 1715-1724.

        Williams' correction:

            n(t) = 2.107 * 10^7 * (t + 2754)^-1.526

        Williams, A. N. (2012). The use of summed radiocarbon probability
        distributions in archaeology: a review of methods.
        Journal of Archaeological Science, 39(3), 578-589.

        The taphonomic correction is not applied homogeneously, but taking into
        account the percentage of open sites, rock shelters and caves samples
        in each age. This percentage correction is given by the vector
        "factor". If factor=None, then correction is applied homogeneously.

        Arguments:
            factor -- Percentage of correction between 0 and 1.
        """

        corr = self.copy()

        if function == "Surovell":
            taph = 5726442 * np.power((self.T[0] + 2176.4), -1.3925309)
        elif function == "Williams":
            taph = 21070000 * np.power((self.T[0] + 2754), -1.526)
        else:
            print("ERROR: Invalid argument for 'function' variable")
            sys.exit()

        corr_y = [c / t if t > 0 else c for c, t in zip(np.array(self.T[1]), taph)]
        corr_y = np.array(corr_y) * np.sum(self.T[1]) / np.sum(corr_y)

        corr.T[1] = corr_y * factor + (1.0 - factor) * self.T[1]

        return corr

    def rollmean(self, m):
        """Performs a Rolling Mean of a 2*m+1 window (-m to +m)

        Note:
         - Returns a SPD of smaller size.
         - Variable m should always be greater than 0

        """

        a = self.copy()
        n = 2 * m + 1
        ret = np.cumsum(a.T[1], dtype=float)
        ret[n:] = ret[n:] - ret[:-n]
        a.T[1][m:-m] = ret[n - 1 :] / n

        return a[m:-m]

    def simdates(self, N, calcurve_name, c14std, seed=0):
        """Simulated dates drawn from SPD curve

        Performs a Bootstrap reseampling on the original SPD and generates a
        new list of simulated dates after backcalibrating. The errors are
        assigned at random from the original distribution.

        Arguments:
            N -- Number of dates to simulate
            seed -- Seed for random generator (if any)
            calcurve_name -- Calibration curve name for backcalibrating
            c14std -- List with all original errors for generating the simulated
                      std.
        """

        np.random.seed = seed
        SPDcum = np.cumsum(self.T[1])

        calcurve = CalibrationCurve(calcurve_name)

        dates = []
        for i in range(N):
            ran = np.random.random()
            idx = (np.abs(SPDcum - ran)).argmin()
            date = self[idx][0]

            idx2 = (np.abs(calcurve.T[0] - date)).argmin()
            c14date = calcurve[idx2][1]

            std = np.random.choice(c14std)
            idname = "sim_" + str(i)
            dates.append([idname, c14date, std])

        return dates

    def simSPD(self, calcurve_name, c14std, seed=0):
        """Simulated SPD curve generated from the original SPD

        Performs a Bootstrap reseampling on the original SPD and generates a
        new list of simulated dates that are backcalibrated in the 14C axis,
        then those dates are calibrated again for constructing the SPD. The
        number of simulated dates is the same than the original curve.

        Arguments:
            calcurve_name -- Calibration curve name for backcalibrating
            c14std -- List with all original errors for generating the simulated
                      std.
            seed -- Seed for random generator (if any)
        """

        simdateslist = self.simdates(self.ndates, calcurve_name, c14std, seed)
        simdates = [RadiocarbonDetermination(x[1], x[2], x[0]) for x in simdateslist]

        # Calibrated date-by-date in order to save memory
        caldates = []
        for x in simdates:
            cal_r = x.calibrate(calcurve_name, norm=True)
            del cal_r.calibration_curve
            del cal_r.radiocarbon_sample
            del cal_r.intervals
            del cal_r.median
            caldates.append(cal_r)

        simulatedSPD = SPD(caldates)

        return simulatedSPD


class FreqHist(np.ndarray):
    """A Frequency Histogram of calibrated dates"""

    def __new__(cls, caldates, bins):
        """Initializes the FreqHist class from calibrated dates

        Arguments:
            caldates -- List with all calibrated dates (after R.calibrate()) in a list
            binwidth -- Array with the bins

        """

        meds = [x.median for x in caldates]

        freqs, bins = np.histogram(meds, bins)

        avbins = 0.5 * (bins[1:] + bins[:-1])

        obj = np.asarray(np.column_stack((avbins, freqs))).view(cls)

        return obj

    def taphcorr(self, function="Surovell", factor=None):
        """Performs a taphonomic correction over the Frequency Histogram

        Taphonomic Correction can be made according two different curves
        indicated by the variable 'function': Surovell's or Williams' correction:

        Surovell's correction:

            n(t) = 5.726442 * 10^6 * (t + 2176.4)^-1.3925309

        Surovell, T. A., Finley, J. B., Smith, G. M., Brantingham, P. J.,
        & Kelly, R. (2009). Correcting temporal frequency distributions for
        taphonomic bias. Journal of Archaeological Science, 36(8), 1715-1724.

        Williams' correction:

            n(t) = 2.107 * 10^7 * (t + 2754)^-1.526

        Williams, A. N. (2012). The use of summed radiocarbon probability
        distributions in archaeology: a review of methods.
        Journal of Archaeological Science, 39(3), 578-589.

        The taphonomic correction is not applied homogeneously, but taking into
        account the percentage of open sites, rock shelters and caves samples
        in each age. This percentage correction is given by the vector
        "factor". If factor=None, then correction is applied homogeneously.

        Arguments:
            factor -- Percentage of correction in each bin.
        """

        corr = self.copy()

        if function == "Surovell":
            taph = 5726442 * np.power((self.T[0] + 2176.4), -1.3925309)
        elif function == "Williams":
            taph = 21070000 * np.power((self.T[0] + 2754), -1.526)
        else:
            print("ERROR: Invalid argument for 'function' variable")
            sys.exit()

        if factor is None:
            taph = np.array(taph)
        else:
            taph = np.array(taph) * factor

        corr_y = [c / t if t > 0 else c for c, t in zip(np.array(self.T[1]), taph)]
        corr_y = np.array(corr_y) * np.sum(self.T[1]) / np.sum(corr_y)

        corr.T[1] = corr_y

        return corr



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# filename: core.py
# Copyright 2020 Stefano Costa <steko@iosa.it>
# Copyright 2017 Mario Gutiérrez-Roig <mariogutierrezroig@gmail.com>
# Copyright 2022 Hythem Sidky <sidky@outlook.com>
#
# This file is part of IOSACal, the IOSA Radiocarbon Calibration Library.

# IOSACal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# IOSACal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with IOSACal.  If not, see <http://www.gnu.org/licenses/>.

import importlib.resources
from math import exp, pow, sqrt

import numpy as np
from scipy import stats

from iosacal.hpd import hpd_interval


def calibrate(f_m, sigma_m, f_t, sigma_t):
    r"""Calibration formula as defined by Bronk Ramsey 2008.

            .. math::

               P(t) \propto \frac{\exp \left[-\frac{(f_m - f(t))^2}{2 (\sigma^2_{fm} + \sigma^2_{f}(t))}\right]}{\sqrt{\sigma^2_{fm} + \sigma^2_{f}(t)}}

    See doi: 10.1111/j.1475-4754.2008.00394.x for a detailed account."""

    sigma_sum = pow(sigma_m, 2) + pow(sigma_t, 2)
    P_t = exp(-pow(f_m - f_t, 2) / (2 * (sigma_sum))) / sqrt(sigma_sum)
    return P_t


def confint(boot_set):
    """Calculates the Confidence Intervals of a Bootstrap set.

    This function returns the 68.27% C.I. (1-sigma) and 95.44% C.I. (2-sigma)
    of a set of simulated curves.

    Arguments:
        boot_set -- List with simulated curves as (x,y).

    """

    mini = np.int(np.min([np.min(x.T[0]) for x in boot_set]))
    maxi = np.int(np.max([np.max(x.T[0]) for x in boot_set]))
    Nboot = len(boot_set)

    yrange = []
    for spd in boot_set:
        y = np.array(
            [[val[0], val[1]] for val in spd if (val[0] >= mini) and (val[0] <= maxi)]
        )
        yy = np.lib.pad(
            y.T[1],
            (np.int(y.T[0][0] - mini), np.int(maxi - y.T[0][-1])),
            "constant",
            constant_values=0,
        )
        yrange.append(yy)

    yrange = np.array(yrange).T

    CIxrange = np.arange(mini, maxi + 1)
    CI95sup = []
    CI68sup = []
    CImed = []
    CI68inf = []
    CI95inf = []
    for val in yrange:
        CI95sup.append(np.sort(val)[np.int(0.9772 * Nboot)])
        CI68sup.append(np.sort(val)[np.int(0.8413 * Nboot)])
        CImed.append(np.sort(val)[np.int(0.5 * Nboot)])
        CI68inf.append(np.sort(val)[np.int(0.1587 * Nboot)])
        CI95inf.append(np.sort(val)[np.int(0.0228 * Nboot)])

    return np.array([CIxrange, CI95inf, CI68inf, CImed, CI68sup, CI95sup])


class CalibrationCurve(np.ndarray):
    """A radiocarbon calibration curve.

    Calibration data is loaded at runtime from source data files, and
    exposed a ``numpy.ndarray`` object.

    The curve can be loaded from:

    - a file if a path is passed as argument e.g. mycurve.14c
    - one of the standard curves such as "intcal20" or "marine20"

    Implementation from
    http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

    """

    def __new__(cls, curve):
        try:
            curve.shape and curve.size and curve.ndim and curve.title
        except AttributeError:
            try:
                with open(curve, encoding="latin-1") as curve_file:
                    title = curve_file.readline().strip("#\n")
            except FileNotFoundError:
                try:
                    curve_resource_filename = (
                        importlib.resources.files("iosacal.data") / f"{curve}.14c"
                    )
                    with open(
                        curve_resource_filename, encoding="latin-1"
                    ) as curve_file:
                        title = curve_file.readline().strip("#\n")
                except FileNotFoundError:
                    available = importlib.resources.files("iosacal.data").iterdir()
                    # print(list(available))
                    message = ""
                    message = "\n".join(
                        [
                            res.name[:-4]
                            for res in available
                            if res.name.endswith(".14c")
                        ]
                    )
                    message = "Available calibration curves: \n" + message
                    raise FileNotFoundError(message)
                else:
                    curve_filename = curve_resource_filename
            else:
                curve_filename = curve
            title = open(curve_filename, encoding="latin-1").readline().strip("#\n")
            _genfrom = np.genfromtxt(curve_filename, delimiter=",", encoding="latin-1")
            # linear interpolation
            ud_curve = np.flipud(_genfrom)  # the sequence must be *increasing*
            curve_arange = np.arange(ud_curve[0, 0], ud_curve[-1, 0], 1)
            values_interp = np.interp(curve_arange, ud_curve[:, 0], ud_curve[:, 1])
            stderr_interp = np.interp(curve_arange, ud_curve[:, 0], ud_curve[:, 2])
            ud_curve_interp = np.array(
                [curve_arange, values_interp, stderr_interp]
            ).transpose()
            _darray = np.flipud(ud_curve_interp)  # back to *decreasing* sequence
            # We cast _darray to be our class type
            obj = np.asarray(_darray).view(cls)
            # add the new attribute to the created instance
            obj.title = title
            # Finally, we must return the newly created object:
            return obj
        else:
            return curve

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
        self.title = getattr(obj, "title", None)

    def mixing(self, curve2_name, P, D=0, deltaR=0, err_deltaR=0):
        """Transforms the calibration curve by mixing with another curve.

        The mathematical equation for the mixing is the same than in
        OxCal, as indicated in:

        http://c14.arch.ox.ac.uk/oxcal3/math_ca.htm#mix_curves

        Curve 1 (self): R_1 +/- E_1
        Curve 2 : R_2 +/- E_2
        Mixing : R_m +/- E_m

        where,

        R_m = (1 - P) * R_1 + P * R_2
        E_m = sqrt(((1 - P) * E_1)^2 + (P * E_2)^2 + (D * (R_1 - R_2))^2)

        The local corrections fo the reservoir effect are applicated over the
        calibration curve as:

        R_2_corrected(t) = R_2(t) + deltaR
        E_2_corrected(t) = sqrt((E_2(t))^2 + err_deltaR^2)

        Arguments:
            curve2_name -- The name of the second curve to mix
            P -- Proportion of the second curve
            D -- Error of the proportion
            deltaR -- Reservoir Effect
            err_deltaR -- Error in deltaR

        """
        curve1 = self

        curve2 = CalibrationCurve(curve2_name)

        # Reservoir Effects
        if deltaR > 0:
            curve2.T[1] += deltaR
            curve2.T[2] = np.sqrt(np.power(curve2.T[2], 2) + np.power(err_deltaR, 2))

        self.T[0] = curve1.T[0]
        self.T[1] = (1.0 - P) * curve1.T[1] + P * curve2.T[1]
        self.T[2] = np.sqrt(
            np.power((1.0 - P) * curve1.T[2], 2)
            + np.power(P * curve2.T[2], 2)
            + np.power(D * (curve1.T[1] - curve2.T[1]), 2)
        )

        # add the new attribute to the created instance
        self.title = "Mixed_curve"

    def __str__(self):
        return "CalibrationCurve( %s )" % self.title


class RadiocarbonDetermination(object):
    """A radiocarbon determination as reported by the lab."""

    def __init__(self, date, sigma, id):
        try:
            self.date = int(date)
        except ValueError as e:
            raise ValueError(
                "the value {} could not be read as an integer".format(date)
            )
        self.sigma = sigma
        self.id = id

    def calibrate(self, curve, norm=True, cutoff=5):
        """Perform calibration, given a calibration curve.

        Arguments:
            curve -- Calibration curve as file path or short name e.g. "intcal20"
            norm -- Normalization in Calendar Scale
            cutoff -- How much of radiocarbon gaussian range are we considering (in sigmas)

        """

        curve = CalibrationCurve(curve)

        _calibrated_list = []

        # We apply the cutoff for the gaussian range in 14C scale
        idx = np.where(
            (curve.T[1] > self.date - cutoff * self.sigma)
            & (curve.T[1] < self.date + cutoff * self.sigma)
        )
        idxmin = np.min(idx)
        idxmax = np.max(idx)

        for i in curve[idxmin:idxmax]:
            f_t, sigma_t = i[1:3]
            ca = calibrate(self.date, self.sigma, f_t, sigma_t)
            _calibrated_list.append((i[0], ca))

        if norm == True:
            x = np.array(_calibrated_list).T[0]
            y = np.array(_calibrated_list).T[1] / np.sum(
                np.array(_calibrated_list).T[1]
            )
            calibrated_curve = np.column_stack((x, y))
        else:
            calibrated_curve = np.array(_calibrated_list)

        cal_age = CalAge(calibrated_curve, self, curve)
        return cal_age

    def __str__(self):
        return "RadiocarbonDetermination( {id} : {date:.0f} ± {sigma:.0f} )".format(
            **self.__dict__
        )


class R(RadiocarbonDetermination):
    """Shorthand for RadiocarbonDetermination."""

    pass


class CalAge(np.ndarray):
    """A calibrated radiocarbon age.

    It is expressed as a probability distribution on the calBP
    calendar scale.

    Implementation from
    http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

    """

    def __new__(cls, input_array, radiocarbon_sample, calibration_curve):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj.radiocarbon_sample = radiocarbon_sample
        obj.calibration_curve = calibration_curve
        obj.intervals = {68: hpd_interval(obj, 0.318), 95: hpd_interval(obj, 0.046)}
        obj.median = np.mean(hpd_interval(obj, 0.5))
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
        self.radiocarbon_sample = getattr(obj, "radiocarbon_sample", None)
        self.calibration_curve = getattr(obj, "calibration_curve", None)

    def calendar(self):
        """Return the calibrated age on the calAD calendar scale.

        This method returns a copy of the calBP array, leaving the
        main object untouched.

        """

        calendarray = self.copy()
        calendarray[:, 0] *= -1
        calendarray[:, 0] += 1950
        return calendarray

    def quantiles(self):
        """Return the median, one, and two sigma ages from a RadiocarbonDetermination object."""
        quantiles = {}

        # get cumulative sum of rows
        cumsum = self[:, 1].cumsum()

        # median is the row whose cumsum is closest to 0.5
        quantiles[50] = self[np.argmin(abs(cumsum - 0.5)), 0]

        # Add 1 & 2 sigma +/- to quantiles dictionary
        for q in [68, 95]:
            quantiles[q] = [
                self.intervals[q][-1].to_year,
                self.intervals[q][0].from_year,
            ]

        return quantiles

    def __str__(self):
        return 'CalAge( {radiocarbon_sample.id} based on "{calibration_curve.title}" )'.format(
            **self.__dict__
        )


def combine(determinations):
    """Combine n>1 determinations related to the same event.

    ``determinations`` is an iterable of tuples (mean, error).

    This covers case 1 as described by Ward and Wilson in their
    seminal 1978 paper (DOI: 10.1111/j.1475-4754.1978.tb00208.x)

    """

    m, s, ids = zip(*[(d.date, d.sigma, d.id) for d in determinations])

    # pooled mean
    pool_m = sum(mi / si**2 for mi, si in zip(m, s)) / sum(1 / si**2 for si in s)

    # standard error on the pooled mean
    pool_s = sqrt(1 / sum(1 / si**2 for si in s))

    # test statistic
    test = sum((mi - pool_m) ** 2 / si**2 for mi, si in zip(m, s))

    desc = "Combined from {} with test statistic {:.3f}".format(", ".join(ids), test)

    return R(pool_m, pool_s, desc)


def simulate(calendar_age, sigma, id, curve, BP="bp"):
    """Simulate a radiocarbon determination for a sample of a known date.

    This function simulates a radiocarbon determination from a lab for
    a sample of a known calendar age with an specified lab error.

    Parameters:
    calendar_age (int): Sample calendar age to be simulated.
    sigma (int): Expected measurement error from radiocarbon lab.
    id (str): ID of the sample.
    curve (str): Calibration curve as file path or short name e.g. "intcal20".
    BP (str): Specifies Gregorian (AD/CE) or BP date for `calendar_age` input.

    Returns:
    RadiocarbonDetermination: Simulated radiocarbon determination as reported by a lab.

    """

    if BP.lower() not in ["ce", "ad", "bp"]:
        raise ValueError(
            f"Specified date must be one of (CE, AD, BP). You provided {BP}."
        )

    curve = CalibrationCurve(curve)

    # Convert to BP if applicable.
    if BP.lower() in ["ce", "ad"]:
        calendar_age = 1950 - calendar_age

    # Get matching mean 14C age and uncertainty from calibration curve.
    idx = np.argwhere(curve[:, 0] == calendar_age).item()
    uncalibrated_mean, uncalibrated_std = curve[idx, 1:]

    # calculate combined std.
    combined_std = sqrt(uncalibrated_std**2 + sigma**2)

    date = stats.norm.rvs(loc=uncalibrated_mean, scale=combined_std).astype(int)

    return R(date, sigma, id)



# -*- coding: utf-8 -*-
# filename: text.py
# Copyright 2018 Stefano Costa <steko@iosa.it>
#
# This file is part of IOSACal, the IOSA Radiocarbon Calibration Library.

# IOSACal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# IOSACal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with IOSACal.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import indent

import iosacal


def single_text(calibrated_age, BP="bp"):
    """Output calibrated age as simple Markdown text to the terminal."""

    formatted_intervals = dict()
    for a, i in calibrated_age.intervals.items():
        formatted_intervals[a] = indent("{:{fmt}}".format(i, fmt=BP), "* ")

    output = """
# {0.radiocarbon_sample.id}

Calibration of {0.radiocarbon_sample.id}: {0.radiocarbon_sample.date} ± {0.radiocarbon_sample.sigma} BP

## Calibrated age

{0.calibration_curve.title}

### 68.2% probability

{1[68]}

### 95.4% probability

{1[95]}

----

IOSACal v{2}

""".format(calibrated_age, formatted_intervals, iosacal.__VERSION__)

    return output



from iosacal.core import (
    calibrate,
    combine,
    simulate,
    R,
    CalibrationCurve,
    RadiocarbonDetermination,
)

import importlib.resources

import numpy as np
from scipy import stats

import pytest


class TestCalibrate:
    def test_calibrate(self):
        assert calibrate(7345, 34, 7513.0, 13.0) == 6.50743743801389e-07
        assert calibrate(7345, 34, 7301.0, 23.0) == 0.013715329452426768

    @pytest.mark.xfail
    def test_error(self):
        with pytest.raises(ValueError):
            calibrate(7345, 0, 7513.0, 10.0)


class TestQuantiles:
    def test_Quantiles(self):
        quantiles = R(3000, 540, "test1").calibrate("intcal20").quantiles()

        print(quantiles)
        assert np.array_equal(quantiles[50], 3190.0)
        assert np.array_equal(quantiles[95], [1828.0, 4572.0])
        assert np.array_equal(quantiles[68], [2469.0, 3844.0])


class TestCalibrationCurve:
    @pytest.fixture(autouse=True)
    def curve(self):
        curve_path = importlib.resources.files("iosacal.data") / "intcal20.14c"
        self.curve = CalibrationCurve(curve_path)

    def test_curve_from_curve(self):
        assert np.array_equal(CalibrationCurve(self.curve), self.curve)

    def test_calibration_curve_title(self):
        assert self.curve.title == "Atmospheric data from Reimer et al (2020)"
        assert np.array_equal(self.curve[0], np.array([54999.0, 50099.05, 1023.7]))
        assert np.array_equal(self.curve[-1], np.array([0.0, 199.0, 11.0]))


class TestCalibrationCurvePath:
    def curve_path(self):
        self.curve = CalibrationCurve("intcal20")
        assert self.curve.title == "Atmospheric data from Reimer et al (2020)"

    def curve_path_error(self):
        with pytest.raises(FileNotFoundError):
            self.curve = CalibrationCurve("non_existing_curve")


class TestRadiocarbonDetermination:
    @pytest.fixture(autouse=True)
    def r(self):
        self.r = RadiocarbonDetermination(7505, 93, "P-769")

    def test_radiocarbon_determination(self):
        assert self.r.date == 7505
        assert self.r.sigma == 93
        assert self.r.id == "P-769"

    def test_radiocarbon_calibration(self):
        r_cal = self.r.calibrate("intcal13")
        assert np.array_equal(
            r_cal[0], np.array([8.96800000e03, 2.3781370616220048e-08])
        )

    def test_radiocarbon_calibration20(self):
        r_cal = self.r.calibrate("intcal20")
        assert np.allclose(r_cal[0], np.array([8.97400000e03, 3.23557224e-08]))


class TestCombine:
    @pytest.fixture(autouse=True)
    def determinations(self):
        self.determinations = (
            R(7345, 35, "Test-A"),
            R(7387, 21, "Test-B"),
            R(7329, 40, "Test-C"),
        )

    def test_combine(self):
        r_combined = combine(self.determinations)
        print(r_combined)
        r_reference = R(
            7367.98,
            16.42,
            "Combined from Test-A, Test-B, Test-C with test statistic 2.201",
        )
        assert r_combined.date == pytest.approx(r_reference.date)
        assert r_combined.sigma == pytest.approx(r_reference.sigma, rel=1e-5)

    def test_error_combine_single(self):
        with pytest.raises(TypeError):
            combine(self.determinations[0])

    def test_error_combine_wrong_datatype(self):
        with pytest.raises(AttributeError):
            combine(
                (
                    self.determinations[0],
                    (7345, 35, "This is a tuple, not a RadiocarbonDetermination"),
                )
            )


class TestSimulate:
    def test_simulate(self):
        curve = CalibrationCurve("intcal20")
        sim = simulate(41250, 23, "simulated", "intcal20")
        real = R(41250, 23, "real")
        assert type(sim) == type(real)

    @pytest.mark.skip(reason="This test does too much and can predictably fail")
    def test_sampling(self):
        curve = CalibrationCurve("intcal20")

        # Test multiple lab errors.
        stds = [0, 11, 21]

        # Index corresponds to 700 CE in intcal20.
        idx = 53749
        assert curve[idx, 0] == 1950 - 700

        curve_mean, curve_std = curve[idx, 1:]

        for std in stds:
            samples = [
                simulate(700, std, "simulated", "intcal20", BP="ce").date
                for _ in range(100)
            ]

            expected_mean = curve_mean
            expected_std = np.sqrt(curve_std**2 + std**2).item()
            test = stats.kstest(
                samples,
                lambda x: stats.norm.cdf(x, loc=expected_mean, scale=expected_std),
            )

            assert test.pvalue > 0.01



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# filename: plot.py
# Copyright 2018 Stefano Costa <steko@iosa.it>
# Copyright 2020 Karl Håkansson <kalle@jomenvisst.de>
# Copyright 2022 Hythem Sidky <sidky@outlook.com>
#
# This file is part of IOSACal, the IOSA Radiocarbon Calibration Library.

# IOSACal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# IOSACal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with IOSACal.  If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
import iosacal
import numpy as np

from scipy.stats import norm

COLORS = {
    "bgcolor": "#e5e4e5",
}


def single_plot(calibrated_age, oxcal=False, output=None, BP="bp"):
    calibrated_age = calibrated_age
    f_m = calibrated_age.radiocarbon_sample.date
    sigma_m = calibrated_age.radiocarbon_sample.sigma
    radiocarbon_sample_id = calibrated_age.radiocarbon_sample.id
    calibration_curve = calibrated_age.calibration_curve
    intervals = calibrated_age.intervals
    sample_interval = np.linspace(f_m + sigma_m * 5, f_m - sigma_m * 5, 100)

    # adjust plot bounds
    min_year, max_year = (50000, -50000)
    min_x = min(calibrated_age[:, 0]) * -1 + 1950
    max_x = max(calibrated_age[:, 0]) * -1 + 1950
    if min_year > min_x:
        min_year = min_x
    if max_year < max_x:
        max_year = max_x

    # do not plot the part of calibration curve that is not visible
    # greatly reduces execution time \o/
    cutmin = calibration_curve[calibration_curve[:, 0] >= (1950 - min_x)]
    cutmax = cutmin[cutmin[:, 0] <= (1950 - max_x)]
    calibration_curve = cutmax

    if BP != "bp":
        if min_year < 0 and max_year > 0:
            ad_bp_label = "BC/AD"
        elif min_year < 0 and max_year < 0:
            ad_bp_label = "BC"
        elif min_year > 0 and max_year > 0:
            ad_bp_label = "AD"
    else:
        ad_bp_label = "BP"

    string68 = "{:{fmt}}".format(calibrated_age.intervals[68], fmt=BP)
    string95 = "{:{fmt}}".format(calibrated_age.intervals[95], fmt=BP)

    fig = plt.figure(figsize=(12, 8))
    fig.clear()
    ax1 = plt.subplot(111)
    ax1.set_facecolor(COLORS["bgcolor"])
    plt.xlabel("Calibrated age ({})".format(ad_bp_label))
    plt.ylabel("Radiocarbon determination (BP)")
    plt.text(
        0.5,
        0.95,
        r"{}: {:.0f} ± {:.0f} BP".format(radiocarbon_sample_id, f_m, sigma_m),
        horizontalalignment="center",
        verticalalignment="center",
        transform=ax1.transAxes,
        bbox=dict(facecolor="white", alpha=0.9, lw=0),
    )
    plt.text(
        0.75,
        0.80,
        "68.2% probability\n{}\n\n95.4% probability\n{}".format(string68, string95),
        horizontalalignment="left",
        verticalalignment="center",
        transform=ax1.transAxes,
        bbox=dict(facecolor="white", alpha=0.9, lw=0),
    )
    plt.text(
        0.0,
        1.0,
        "IOSACal v{}; {}".format(iosacal.__VERSION__, calibration_curve.title),
        horizontalalignment="left",
        verticalalignment="bottom",
        transform=ax1.transAxes,
        size=10,
        bbox=dict(facecolor="white", alpha=0.9, lw=0),
    )

    # Calendar Age

    ax2 = plt.twinx()

    if oxcal is True:
        # imitate OxCal
        ax1.set_facecolor("white")
        ax2.fill_between(
            calibrated_age[:, 0],
            calibrated_age[:, 1] * 0,
            calibrated_age[:, 1],
            facecolor="k",
            alpha=0.3,
            label="Calendar Age",
        )
        ax2.plot(calibrated_age[:, 0], calibrated_age[:, 1], "k", alpha=0)
    else:
        ax2.fill_between(
            calibrated_age[:, 0],
            calibrated_age[:, 1] * 0,
            calibrated_age[:, 1],
            facecolor="k",
            alpha=0.3,
            label="Calendar Age",
        )
        ax2.plot(calibrated_age[:, 0], calibrated_age[:, 1], "k", alpha=0)

    ax2.set_ybound(min(calibrated_age[:, 1]), max(calibrated_age[:, 1]) * 3)
    ax2.set_xbound(min(calibrated_age[:, 0]), max(calibrated_age[:, 0]))
    ax2.set_axis_off()

    # Radiocarbon Age
    sample_curve = norm.pdf(sample_interval, f_m, sigma_m)

    if oxcal is True:
        sample_fill_color = "#fac5cd"
    else:
        sample_fill_color = "w"
    ax3 = plt.twiny(ax1)
    ax3.fill(
        sample_curve, sample_interval, "1.0", facecolor=sample_fill_color, alpha=0.8
    )
    ax3.set_xbound(0, max(sample_curve) * 4)
    ax3.set_axis_off()

    # Calibration Curve

    curve_low = calibration_curve[:, 1] - calibration_curve[:, 2]
    curve_high = calibration_curve[:, 1] + calibration_curve[:, 2]

    ax1.fill_between(
        calibration_curve[:, 0],
        curve_low,
        curve_high,
        facecolor="#000000",
        edgecolor="none",
        alpha=0.15,
    )
    ax1.plot(calibration_curve[:, 0], calibration_curve[:, 1], "k", lw=0.5, alpha=0.7)

    # Confidence intervals

    if oxcal is True:
        for i in intervals[68]:
            ax1.axvspan(
                i.from_year,
                i.to_year,
                ymin=0.05,
                ymax=0.07,
                facecolor="none",
                alpha=0.8,
            )
            ax1.axvspan(
                i.from_year,
                i.to_year,
                ymin=0.068,
                ymax=0.072,
                facecolor="w",
                edgecolor="w",
                lw=2,
            )
        for i in intervals[95]:
            ax1.axvspan(
                i.from_year,
                i.to_year,
                ymin=0.025,
                ymax=0.045,
                facecolor="none",
                alpha=0.8,
            )
            ax1.axvspan(
                i.from_year,
                i.to_year,
                ymin=0.043,
                ymax=0.047,
                facecolor="w",
                edgecolor="w",
                lw=2,
            )
    else:
        for i in intervals[68]:
            ax1.axvspan(
                i.from_year, i.to_year, ymin=0, ymax=0.02, facecolor="k", alpha=0.5
            )
        for i in intervals[95]:
            ax1.axvspan(
                i.from_year, i.to_year, ymin=0, ymax=0.02, facecolor="k", alpha=0.5
            )

    # FIXME the following values 10 and 5 are arbitrary and could be probably
    # drawn from the f_m value itself, while preserving their ratio
    y_lower = np.minimum(f_m - sigma_m * 15, curve_low.min())
    y_upper = np.maximum(f_m + sigma_m * 5, curve_high.max())
    ax1.set_ybound(y_lower, y_upper)
    ax1.set_xbound(min(calibrated_age[:, 0]), max(calibrated_age[:, 0]))

    ax1.invert_xaxis()
    if BP != "bp":
        ax1.set_xticklabels(
            [int(item * -1 + 1950) for item in ax1.get_xticks().tolist()]
        )

    if output:
        plt.savefig(output)


def stacked_plot(
    calibrated_ages, name="Stacked plot", oxcal=False, BP="bp", output=None
):
    """Plot multiple calibrated ages, vertically stacked.

    ``calibrated_ages`` is an iterable of CalAge objects."""

    # Define the legend and descriptive text

    min_year, max_year = (50000, -50000)

    for calibrated_age in calibrated_ages:
        radiocarbon_sample_id = calibrated_age.radiocarbon_sample.id
        calibration_curve = calibrated_age.calibration_curve
        calibration_curve_title = calibration_curve.title
        intervals = calibrated_age.intervals

        min_x = min(calibrated_age[:, 0]) * -1 + 1950
        max_x = max(calibrated_age[:, 0]) * -1 + 1950

        if min_year > min_x:
            min_year = min_x
        if max_year < max_x:
            max_year = max_x

    if BP != "bp":
        if min_year < 0 and max_year > 0:
            ad_bp_label = "BC/AD"
        elif min_year < 0 and max_year < 0:
            ad_bp_label = "BC"
        elif min_year > 0 and max_year > 0:
            ad_bp_label = "AD"
    else:
        ad_bp_label = "BP"

    numrows = len(calibrated_ages)
    fig, axs = plt.subplots(numrows, 1, sharex=True, figsize=(12, 2 * numrows))
    plt.suptitle("{}".format(name))
    if BP == "bp":
        axs[0].invert_xaxis()  # just once, because the axis is shared

    for n, calibrated_age in enumerate(calibrated_ages):
        ax = axs[n]

        # Calendar Age

        ax.fill_between(
            calibrated_age[:, 0],
            calibrated_age[:, 1] * 0,
            calibrated_age[:, 1],
            facecolor="k",
            alpha=0.3,
            label="Calendar Age",
        )
        ax.plot(calibrated_age[:, 0], calibrated_age[:, 1], "k", alpha=0.6)
        ax.set_ybound(min(calibrated_age[:, 1]), max(calibrated_age[:, 1]) * 2)
        ax.set_xbound(min_year, max_year)

        # remove labels for Y axis - values are meaningless
        ax.get_yaxis().set_ticklabels("")

        # Legend
        plt.text(
            0.95,
            0.85,
            "{0.radiocarbon_sample.id!s}".format(calibrated_age),
            horizontalalignment="right",
            verticalalignment="center",
            transform=ax.transAxes,
            bbox=dict(facecolor="white", alpha=0.9, lw=0),
        )

        # Confidence intervals
        for i in calibrated_age.intervals[95]:
            ax.axvspan(
                i.from_year, i.to_year, ymin=0.6, ymax=0.7, facecolor="k", alpha=0.4
            )
        for i in calibrated_age.intervals[68]:
            ax.axvspan(
                i.from_year, i.to_year, ymin=0.6, ymax=0.7, facecolor="k", alpha=1.0
            )
        if n + 1 == numrows:
            if BP != "bp":
                plt.xlabel(
                    "Calibrated age ({})".format(ad_bp_label), y=0.05
                )  # out of plot boundaries :(
            else:
                plt.xlabel("Calibrated age (BP)", y=0.05)

    if output:
        plt.savefig(output)


def iplot(calibrated_ages, **kwds):
    """A generic function for plotting in the IPython Notebook."""

    try:
        # ugly, ugly hack
        calibrated_ages.intervals
    except AttributeError:
        stacked_plot(calibrated_ages, **kwds)
    else:
        single_plot(calibrated_ages, **kwds)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# filename: cli.py
# Copyright 2020 Stefano Costa <steko@iosa.it>
#
# This file is part of IOSACal, the IOSA Radiocarbon Calibration Library.

# IOSACal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# IOSACal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with IOSACal.  If not, see <http://www.gnu.org/licenses/>.


import click

from iosacal import core, plot, text


@click.command()
@click.option(
    "-d",
    "--date",
    help="non calibrated radiocarbon BP date for sample",
    required=True,
    type=int,
    multiple=True,
)
@click.option(
    "-s",
    "--sigma",
    help="standard deviation for date",
    required=True,
    type=int,
    multiple=True,
)
@click.option(
    "--id", type=str, multiple=True, required=True, help="sample identification"
)
@click.option(
    "-p",
    "--plot",
    "plot_",
    type=bool,
    default=False,
    is_flag=True,
    help="output results to graphic plot",
)
@click.option(
    "-c",
    "--curve",
    type=str,
    default="intcal20",
    help="calibration curve (either as shortname or path to file)",
)
@click.option(
    "-o", "--oxcal", type=bool, default=False, help="draw plots more OxCal-like looking"
)
@click.option(
    "-n", "--name", type=click.Path(), default="iosacal", help="name of output plot"
)
@click.option(
    "-1",
    "--single/--no-single",
    type=bool,
    default=True,
    help="generate single plots for each sample",
)
@click.option(
    "-m",
    "--stacked",
    type=bool,
    default=False,
    is_flag=True,
    help="generate stacked plot with all samples",
)
@click.option(
    "--bp",
    "BP",
    flag_value="bp",
    default=True,
    help="express dates in Calibrated BP Age",
)
@click.option(
    "--ad", "BP", flag_value="ad", help="express dates in Calibrated BC/AD Calendar Age"
)
@click.option(
    "--ce",
    "BP",
    flag_value="ce",
    help="express dates in Calibrated BCE/CE Calendar Age",
)
def main(date, sigma, id, plot_, curve, oxcal, name, single, stacked, BP):
    """Main program procedure.

    By default produces text output to stdout for each sample."""

    if not (len(date) == len(sigma) == len(id)):
        raise click.BadParameter(
            "Number of dates, sigma and id must be equal",
            param_hint=["date", "sigma", "id"],
        )

    curve = core.CalibrationCurve(curve)
    calibrated_ages = []
    for d, s, id in zip(date, sigma, id):
        rs = core.R(d, s, id)
        ca = rs.calibrate(curve)
        calibrated_ages.append(ca)
        if plot_ and single is True:
            outputname = "{}_{:d}_{:d}.pdf".format(id, d, s)
            plot.single_plot(ca, oxcal=oxcal, output=outputname, BP=BP)
        else:
            click.echo(text.single_text(ca, BP))
    if plot_ and stacked is True:
        plot.stacked_plot(
            calibrated_ages,
            oxcal=oxcal,
            name=name,
            output="{}.pdf".format(name),
        )



from click.testing import CliRunner
from iosacal.cli import main


def test_iosacal_cli():
    runner = CliRunner()
    result = runner.invoke(main, ["-d", "1100", "-s", "45", "--id", "TEST"])
    assert result.exit_code == 0
    assert "iosacal" in result.output.lower()


def test_iosacal_missing_option():
    runner = CliRunner()
    result = runner.invoke(main, ["-d", "1100", "-s", "45"])
    assert result.exit_code != 0


def test_iosacal_equal_number():
    runner = CliRunner()
    result = runner.invoke(
        main, ["-d", "1100", "-s", "45", "--id", "TEST", "-d", "1080"]
    )
    assert result.exit_code != 0
    assert "equal" in result.output


def test_iosacal_plot():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(main, ["-d", "1100", "-s", "45", "--id", "TEST", "-p"])
        assert result.exit_code == 0


def test_iosacal_plot_stacked():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            main,
            [
                "-d",
                "1100",
                "-s",
                "45",
                "--id",
                "TEST1",
                "-d",
                "1080",
                "-s",
                "35",
                "--id",
                "TEST2",
                "-p",
                "-m",
            ],
        )
        assert result.exit_code == 0



import pytest
import iosacal.plot as plot

from iosacal.core import R, CalibrationCurve


class TestSinglePlot:
    @pytest.fixture(autouse=True)
    def determination(self):
        determination = R(7345, 32, "Test-A")
        self.cal_age = determination.calibrate(CalibrationCurve("intcal13"))

    @pytest.mark.mpl_image_compare
    def test_succeeds(self):
        fig = plot.single_plot(self.cal_age)


class TestMultiPlot:
    @pytest.fixture(autouse=True)
    def determinations(self):
        determinations = (
            R(7345, 35, "Test-A"),
            R(7387, 21, "Test-B"),
            R(7329, 40, "Test-C"),
        )
        self.cal_ages = [
            determination.calibrate("intcal20") for determination in determinations
        ]

    @pytest.mark.mpl_image_compare
    def test_succeeds(self):
        fig = plot.stacked_plot(self.cal_ages)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# filename: hpd.py
# Copyright 2008-2009 Stefano Costa <steko@iosa.it>
# Copyright 2008 David Laban <alsuren@gmail.com>
#
# This file is part of IOSACal, the IOSA Radiocarbon Calibration Library.

# IOSACal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# IOSACal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with IOSACal.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
from copy import copy
from numpy import asarray


def findsorted(n, array):
    """Return sorted array and index of n inside array."""
    a = asarray(array)
    a.sort()
    i = a.searchsorted(n)
    return a, i


def prev(n, array):
    """Find interval between n and its previous, inside array."""

    a, i = findsorted(n, array)
    if i - 1 < 0:
        prev = None
    else:
        prev = a[i - 1]
    return prev


def next(n, array):
    """Find interval between n and its next, inside array."""

    a, i = findsorted(n, array)
    try:
        next = a[i + 1]
    except IndexError:
        next = None
    return next


def alsuren_hpd(calibrated_age, alpha):
    """Return year spans that have the required Highest Probability Density."""
    hpd_curve = calibrated_age.copy()
    # sort rows by second column in inverse order
    hpd_sorted = hpd_curve[hpd_curve[:, 1].argsort(),][::-1]
    hpd_cumsum = hpd_sorted[:, 1].cumsum()
    # normalised values
    hpd_cumsum /= hpd_cumsum[-1]

    threshold_index = hpd_cumsum.searchsorted(1 - alpha)
    threshold_p = hpd_sorted[threshold_index][1]
    threshold_index = calibrated_age[:, 1] > threshold_p
    hpd = list(hpd_curve[threshold_index, 0])

    confidence_intervals = list()

    for i in hpd:
        # ^ is the XOR operator
        if (prev(i, hpd_curve[:, 0]) not in hpd) ^ (
            next(i, hpd_curve[:, 0]) not in hpd
        ):
            confidence_intervals.append(i)
    return asarray(confidence_intervals).reshape(len(confidence_intervals) // 2, 2)


def confidence_percent(years, calibrated_age):
    """Return HPD as percent value for a given span of years."""
    percent_curve = calibrated_age.copy()
    percent_curve[:, 1] /= percent_curve[:, 1].sum()
    percent_sorted = percent_curve[percent_curve[:, 0].argsort(),]

    year1_index = percent_sorted[:, 0].searchsorted([years[0]])
    year2_index = percent_sorted[:, 0].searchsorted([years[1]])
    indices = [percent_sorted[:, 0].searchsorted([year]) for year in years]
    indices.sort()
    min_year = indices[0][0]
    max_year = indices[1][0]

    confidence_interval = percent_sorted[min_year : max_year + 1, 1]
    percent_result = confidence_interval.sum()
    return float(percent_result)


class ConfIntv(namedtuple("ConfIntv", ["from_year", "to_year", "conf_perc"])):
    __slots__ = ()

    def __format__(self, fmt="bp"):
        def ad_bc_prefix(year, prefixes="ad"):
            """Return a string with BC/AD prefix and the given year."""
            if prefixes == "ad":
                neg, pos = ("BC", "AD")
            elif prefixes == "ce":
                neg, pos = ("BCE", "CE")
            if year < 0:
                yearf = "{0} {1:.0f}".format(neg, abs(year))
            else:
                yearf = "{0} {1:.0f}".format(pos, year)
            return yearf

        fmt = fmt if len(fmt) > 0 else "bp"  # default fmt is ''
        if fmt == "bp":
            f = "{from_year:.0f} CalBP - {to_year:.0f} CalBP ({conf_perc:2.1%})".format(
                **self._asdict()
            )
        elif fmt == "ad" or "ce":
            ad = {
                "from_year": ad_bc_prefix(1950 - self.from_year, fmt),
                "to_year": ad_bc_prefix(1950 - self.to_year, fmt),
                "conf_perc": self.conf_perc,
            }
            f = "{from_year} - {to_year} ({conf_perc:2.1%})".format(**ad)
        return f

    __str__ = __format__


class ConfIntvList(list):
    def __format__(self, fmt):
        return "\n".join(ci.__format__(fmt) for ci in self)

    __str__ = __format__


def hpd_interval(calibrated_age, alpha):
    """Wrapper around other functions, returns a single object."""

    res = ConfIntvList()
    intervals = alsuren_hpd(calibrated_age, alpha)
    for interval in intervals:
        percent = confidence_percent(interval, calibrated_age)
        res.append(ConfIntv(interval[0], interval[1], percent))
    return res
